#!/home/vagrant/janus/venv/bin/python3
"""Key Operations

This module is ment to connect OpenABE to the rest of the project
throught a REST API.

Example:
    $ python key_operations.py --cli /home/vagrant/janus/cli/ --host 0.0.0.0 \
      --port 8080

Attributes:
    cli  (int):
    host (str):
    port (int):

TODO:
    US.TIK.1-3
    US.CIK.1-4

Sources:
https://flask.palletsprojects.com/en/2.0.x/patterns/fileuploads/
"""
import os
import time
import argparse
import logging
import subprocess
import threading
import shutil
import random
import json
import string
import common
from flask import Flask, request, jsonify, make_response, flash, url_for, send_from_directory
from werkzeug.utils import secure_filename
from functools import wraps

CLI_PATH = ''
HOST_IP = ''
PORT = ''
UPLOAD_FOLDER = common.UPLOAD_FOLDER
ALLOWED_EXTENSIONS = common.ALLOWED_EXTENSIONS

logging.basicConfig(format='%(asctime)s %(message)s')
logger = logging.getLogger('key_operations')

def check_auth(usr, password):
    # TODO: link auth to OpenLdap
    return usr == 'test' and password == 'password'

def login_required(f):
    @wraps(f)
    def wrapped_view(**kwargs):
        auth = request.authorization
        if not (auth and check_auth(auth.username, auth.password)):
            return ('Unauthorized', 401, {
                'WWW-Authenticate': 'Basic realm="Login Required"'
            })
        return f(**kwargs)
    return wrapped_view

def default_success_code():
    message = {
        'message': 'OK',
        'status_code': 200
    }
    return make_response(jsonify(message), 200)

def default_error_code(code, stdout, stderr):
    message = {
        'message': 'FAIL',
        'stdout': stdout,
        'stderr': stderr,
        'status_code': code
    }
    return make_response(jsonify(message), code)

def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS

def subprocess_and_call(on_exit, on_fail, list_args):
    thread = threading.Thread(target=subprocess_and_wait, args=(on_exit, on_fail, list_args))
    thread.start()
    return thread

def subprocess_and_wait(on_exit, on_fail, list_args):
    proc = subprocess.Popen(' '.join(list_args), shell=True)
    proc.wait()
    (stdout, stderr) = proc.communicate()
    if proc.returncode != 0:
        return on_fail(500, stdout, stderr)
    else:
        return on_exit()

app = Flask(__name__)
app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
app.config['MAX_CONTENT_LENGTH'] = 16 * 1000 * 1000 #16 MB

@app.route("/")
def hello_world():
    return "<p>Hello World<p>"

@app.route('/file', methods=['GET', 'POST'])
def upload_file():
    if request.method == 'POST':
        # check if the post request has the file part
        if 'file' not in request.files:
            flash('No file part')
            return redirect(request.url)
        file = request.files['file']
        # If the user does not select a file, the browser submits an
        # empty file without a filename.
        if file.filename == '':
            flash('No selected file')
            return redirect(request.url)
        if file and allowed_file(file.filename):
            filename = secure_filename(file.filename)
            file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
            return redirect(url_for('download_file', name=filename))
    return '''
    <!doctype html>
    <title>Upload new File</title>
    <h1>Upload new File</h1>
    <form method=post enctype=multipart/form-data>
      <input type=file name=file>
      <input type=submit value=Upload>
    </form>
    '''

@app.route('/uploads/<name>')
def download_file(name):
    return send_from_directory(app.config["UPLOAD_FOLDER"], name)

@app.route("/create_org/<org>", methods=['POST'])
@login_required
def create_org(org):
    """Generates MSK, MPK, and Prepares MPK for download
    """
    # TODO: Admin use only
    if os.path.exists('{0}.msk.kpabe'.format(org)):
        return default_error_code(500, '', 'Keys Exists')
    resp = subprocess_and_wait(default_success_code,
                               default_error_code,
                               [os.path.join(CLI_PATH, 'oabe_setup'), '-s KP -p', org])
    if resp.status_code != 200:
        return resp
    try:
        shutil.copy2('{0}.mpk.kpabe'.format(org), os.path.join(UPLOAD_FOLDER, '{0}.mpk.kpabe'.format(org)))
    except e:
        return default_error_code(500, 'File Copy Error', e)
    return make_response(jsonify(message={'message': 'OK',
                                          'files': ['http://{0}:{1}/uploads/{2}.mpk.kpabe'.format(HOST_IP, PORT, org)],
                                          'status_code': 200
                                         }), 200)

@app.route("/mpk/<org>", methods=['GET'])
@login_required
def get_mpk(org):
    """Prepares MPK for download
    """
    if not os.path.exists('{0}.mpk.kpabe'.format(org)):
        return default_error_code(500, '', 'Keys Does Not Exists')
    try:
        shutil.copy2('{0}.mpk.kpabe'.format(org), os.path.join(UPLOAD_FOLDER, '{0}.mpk.kpabe'.format(org)))
    except e:
        return default_error_code(500, 'File Copy Error', e)
    return make_response(jsonify(message={'message': 'OK',
                                          'files': ['http://{0}:{1}/uploads/{2}.mpk.kpabe'.format(HOST_IP, PORT, org)],
                                          'status_code': 200
                                         }), 200)

@app.route("/get_key/<org>/<key_name>", methods=['POST'])
@login_required
def get_key(org, key_name):
    # TODO: Admin Use Only
    # JSON content KP-ABE policy
    data = request.get_json()
    if data is None:
        return default_error_code(400, 'application/json', 'json content empty')
    policy = data.get('policy', None)
    if policy is None:
        return default_error_code(400, 'policy="policy_required', 'Client JSON Error')
    if not os.path.exists('{0}.msk.kpabe'.format(org)):
        return default_error_code(400, '', 'Org does not exist')
    resp = subprocess_and_wait(default_success_code,
                               default_error_code,
                               [os.path.join(CLI_PATH, 'oabe_keygen'),
                                '-s KP -p', org,
                                '-i', '\"{0}\"'.format(policy),
                                '-o', key_name])
    if resp.status_code != 200:
        return resp
    try:
        shutil.copy2('{0}.key'.format(key_name), os.path.join(UPLOAD_FOLDER, '{0}.key'.format(key_name)))
    except e:
        return default_error_code(500, 'File Copy Error', e)
    return make_response(jsonify(message={'message': 'OK',
                                          'files': ['http://{0}:{1}/uploads/{2}.key'.format(HOST_IP, PORT, key_name)],
                                          'status_code': 200
                                         }), 200)

@app.route("/encrypt/")
@login_required
def encrypt():
    # TODO: Admin Use Only
    # JSON content KP-ABE policy
    if 'json' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `json`')
    print(request.files['json'])
    df = request.files['json'].read()
    print(request.files['json'])
    data = json.loads(df)
    if data is None:
        return common.default_error_code(400, 'form-data', 'json content empty')
    org = data.get('org', None)
    if org is None:
        return common.default_error_code(400, 'org required', 'Client JSON Error')
    key = data.get('key', None)
    if key is None:
        return common.default_error_code(400, 'key required', 'Client JSON Error')
    policy = data.get('policy', None)
    if policy is None:
        return default_error_code(400, 'policy="policy_required', 'Client JSON Error')
    # check if the post request has the file part
    if 'file' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `file`')
    file = request.files['file']
    print(file)
    # If the user does not select a file, the browser submits an
    # empty file without a filename.
    if file.filename == '':
        return common.default_error_code(400, 'form-data', 'file content empty in `file`')
    if not os.path.exists('{0}.mpk.kpabe'.format(org)):
        return default_error_code(400, '', 'Org does not exist')
    if not os.path.exists('{0}.key'.format(key)):
        return default_error_code(400, '', 'Org does not exist')

    random_input = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '_input.txt'
    random_output = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '_output.kpabe'
    tmp_input_file = os.path.join(app.config["UPLOAD_FOLDER"], random_input)
    tmp_output_file = os.path.join(app.config["UPLOAD_FOLDER"], random_output)
    try:
        if file and common.allowed_file(file.filename):
            filename = secure_filename(file.filename)
            file.save(tmp_input_file)
        else:
            return default_error_code(500, 'File Save Error', 'filename')
    except:
        return default_error_code(500, 'File Save Error', 'save')
    resp = subprocess_and_wait(default_success_code,
                               default_error_code,
                               [os.path.join(CLI_PATH, 'oabe_enc'),
                                '-s KP -p', org,
                                '-e', '\"{0}\"'.format(policy),
                                '-i', '{0}'.format(tmp_input_file),
                                '-o', tmp_output_file])
    if resp.status_code != 200:
        return default_error_code(500, '', 'Encryption Failure')
    return_resp  = send_from_directory(app.config["UPLOAD_FOLDER"], random_output)
    if os.path.exists(tmp_input_file):
        os.remove(tmp_input_file)
    if os.path.exists(tmp_output_file):
        os.remove(tmp_output_file)
    return return_resp

@app.route("/decrypt/")
@login_required
def decrypt():
    # TODO: Admin Use Only
    # JSON content KP-ABE policy
    if 'json' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `json`')
    data = json.loads(request.files['json'].read())
    if data is None:
        return common.default_error_code(400, 'form-data', 'json content empty')
    org = data.get('org', None)
    if org is None:
        return common.default_error_code(400, 'org required', 'Client JSON Error')
    key = data.get('key', None)
    if key is None:
        return common.default_error_code(400, 'key required', 'Client JSON Error')
    # check if the post request has the file part
    if 'file' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `file`')
    file = request.files['file']
    # If the user does not select a file, the browser submits an
    # empty file without a filename.
    if file.filename == '':
        return common.default_error_code(400, 'form-data', 'file content empty in `file`')
    if not os.path.exists('{0}.mpk.kpabe'.format(org)):
        return default_error_code(400, '', 'Org does not exist')
    if not os.path.exists('{0}.key'.format(key)):
        return default_error_code(400, '', 'key does not exist')

    random_input = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '_input.kpabe'
    random_output = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '_output.txt'
    tmp_input_file = os.path.join(app.config["UPLOAD_FOLDER"], random_input)
    tmp_output_file = os.path.join(app.config["UPLOAD_FOLDER"], random_output)
    try:
        if file and common.allowed_file(file.filename):
            filename = secure_filename(file.filename)
            file.save(tmp_input_file)
        else:
            return default_error_code(500, 'File Save Error', '')
    except:
        return default_error_code(500, 'File Save Error', '')
    print(' '.join([os.path.join(CLI_PATH, 'oabe_dec'),
     '-s KP -p', org,
     '-k', '{0}.key'.format(key),
     '-i', '{0}'.format(tmp_input_file),
     '-o', tmp_output_file]))
    resp = subprocess_and_wait(default_success_code,
                               default_error_code,
                               [os.path.join(CLI_PATH, 'oabe_dec'),
                                '-s KP -p', org,
                                '-k', '{0}.key'.format(key),
                                '-i', '{0}'.format(tmp_input_file),
                                '-o', tmp_output_file])
    if resp.status_code != 200:
        return default_error_code(500, '', 'Decryption Failure')
    return_resp = send_from_directory(app.config["UPLOAD_FOLDER"], random_output)
    if os.path.exists(tmp_input_file):
        os.remove(tmp_input_file)
    if os.path.exists(tmp_output_file):
        os.remove(tmp_output_file)
    return return_resp

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Webserver to handle key operations')
    parser.add_argument('--cli', dest='cli_path', type=str, required=True, help='Path to oabe cli commands')
    parser.add_argument('--host', dest='host_ip', type=str, default='192.168.1.5', help='Flask host IP')
    parser.add_argument('--port', dest='port', type=int, default=8080, help='Flask host port')
    args = parser.parse_args()
    CLI_PATH = args.cli_path
    HOST_IP = args.host_ip
    PORT = args.port
    app.run(host=args.host_ip, port=args.port)
